/* 
 *
 * (c) 2006 Aubit Computing Ltd
 * 
 */

#define EPOCH                                             693608
#define DATE_INVALID                              0
long A4GL_gen_dateno (int day, int month, int year);

static int days_in_month[2][13] = {
  {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
  {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
};

#define leap_year(yr) \
        ((yr) <= 1752 ? !((yr) % 4) : \
        (!((yr) % 4) && ((yr) % 100)) || !((yr) % 400))

/* number of centuries since 1700, not inclusive */
#define centuries_since_1700(yr) \
        ((yr) > 1700 ? (yr) / 100 - 17 : 0)

/* number of centuries since 1700 whose modulo of 400 is 0 */
#define quad_centuries_since_1700(yr) \
        ((yr) > 1600 ? ((yr) - 1600) / 400 : 0)

/* number of leap years between year 1 and this year, not inclusive */
#define leap_years_since_year_1(yr) \
        ((yr) / 4 - centuries_since_1700(yr) + quad_centuries_since_1700(yr))


static int
get_yr (int d)
{
  int e;
  int h, l;
  if (d == DATE_INVALID)
    return d;
  e = (int) ((double) (d - 13 + EPOCH) / 365.2425) + 1;
  h = A4GL_gen_dateno (31, 12, e);
  /*l=gen_dateno(1,1,e); */
  while (1)
    {
      l = h - 365 - leap_year (e) + 1;
      if (d == l || d == h)
	{
	  return e;
	}
      if (d < l)
	{
	  e--;
	  h = l - 1;
	  continue;
	}
      if (d > h)
	{
	  h = h + 365 + leap_year (e);
	  e++;
	  continue;
	}

      break;
    }
  return e;
}


int
A4GL_day_in_year (day, month, year)
     int day, month, year;
{
  int i, leap;

  leap = leap_year (year);
  for (i = 1; i < month; i++)
    day += days_in_month[leap][i];
  return (day);
}


long
A4GL_gen_dateno (int day, int month, int year)
{
  long temp;
  if (month < 1 || month > 12)
    return DATE_INVALID;
  if (day < 1)
    return DATE_INVALID;
  if (day > days_in_month[leap_year (year)][month])
    return DATE_INVALID;
  temp = (long) (year - 1) * 365 + leap_years_since_year_1 (year - 1)
    + A4GL_day_in_year (day, month, year);
  return temp - EPOCH;

}



int
A4GL_get_date (int d, int *day, int *mn, int *yr)
{
  int i, leap;
  int year;
  *day = 0;
  *mn = 0;
  *yr = 0;
  if (d == DATE_INVALID)
    return 0;
  year = get_yr (d);
  *day = d - A4GL_gen_dateno (1, 1, year) + 1;
  leap = leap_year (year);
  for (i = 1; i <= 12; i++)
    {
      *day -= days_in_month[leap][i];
      if (*day <= 0)
	{
	  *day += days_in_month[leap][i];
	  break;
	}
    }
  *mn = i;
  *yr = year;
  return 1;
}
