/* 
 *
 * (c) 2006 Aubit Computing Ltd
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "go.h"
#include "read_db.h"
#include "dbio.h"

FILE *systables = 0;		// File handle for systables
FILE *syscolumns = 0;		// File handle for syscolumns

static void print_progress(int finished) ;
static int needs_quotes (int n);
char *get_col_desc(int coltype,int collength);
extern int show_schema;

static const char *dtparts[]={
		"",
        	"YEAR",
        	"MONTH",
        	"DAY",
        	"HOUR",
        	"MINUTE",
        	"SECOND",
        	"FRACTION"
};


int verbose;
extern char search_tablename[64];
extern char insert_tablename[64];
extern char datefmt_s[10];
extern char *datefmt;
extern char datafile[64];
extern char columnfile[64];
extern char outfile[256];
extern int progress_cnt;
extern int progress_max_cnt;
extern char progress_type;

FILE *out = 0;
extern int clear_first;
extern char dbname[64];
extern char uname[64];
extern char passwd[64];
extern int search_tabid;
extern char mode;
int tabrecord_size = 0;
char *dump_data (char *ptr, int dtype, int len);
char sqlbuff[100000];
void chk_unl(int finished)  ;

#ifdef V7
/* This is how each systables record looks like on disk */
struct s_tabrecord
{
  char tabname[18];
  char owner[8];
  char dirpath[256];
  int tabid;
  short rowsize;
  short ncols;
  short nindexes;
  int nrows;
  int created;
  int version;
  char tabtype[1];
  char audpath[256];
  char stat;
} __attribute__ ((packed));
#else
struct s_tabrecord
{
  char tabname[18];
  char owner[8];
  char dirpath[64];
  int tabid;
  short rowsize;
  short ncols;
  short nindexes;
  int nrows;
  int created;
  int version;
  char tabtype[1];
  char audpath[64];
  char stat;
} __attribute__ ((packed));

#endif


/* Create us one to work with */
struct s_tabrecord tabrecord;


/* This is the column structure on disk... */
struct s_colrecord
{
  char colname[18];
  int tabid;
  short colno;
  short coltype;
  short collength;
  char stat;
} __attribute__ ((packed));



/* This is the column structure we're going to use... */
struct s_column
{
  char colname[19];		// Extra space for the \0
  int tabid;
  short colno;
  short coltype;
  short collength;
  short offset;
};




/* Some debugging routines 

	Note - int values appear to be stored in network short format on disk...
	Note - short values appear to be stored in network short format on disk...

*/

void print_short (char *name, short val)
{
  printf ("%-18s: %d\n", name, a_ntohs (val));
}

void print_double (char *name, double val)
{
  printf ("%-18s: %lf\n", name, val);
}

void print_float (char *name, float val)
{
  printf ("%-18s: %f\n", name, val);
}

void print_hex (char *name, char val)
{
  printf ("%-18s: %x\n", name, val);
}

void print_long (char *name, int val)
{
  printf ("%-18s: %d\n", name, ntohl (val));
}


void openOutFile() {
if (out!=NULL) return;
  if (strlen (outfile) && strcmp (outfile, "-") != 0)
    {
      out = fopen (outfile, "w");
      if (out == 0)
	{
	  fprintf (stderr,"Unable to open output file\n");
	  exit (2);
	}
      if (verbose)
	{
	  printf ("Writing to %s\n", outfile);
	}

    }
  else
    {
      if (verbose)
	{
	  printf ("Writing to stdout\n");
	}
      out = stdout;
    }
}

char *
escape_single_quotes (char *s)
{
  static char *local_s = 0;
  static int local_s_len = 0;
  char *ptr;
  int a;
  int l;
  ptr = strchr (s, '\'');
  if (ptr == 0)
    return s;
  l = 0;
  for (a = 0; a < strlen (s); a++)
    {
      if (s[a] == '\'')
	l++;
      if (s[a] == '\\')
	l++;
      l++;
    }

  if (l > local_s_len)
    {
      local_s_len = l;
      local_s = realloc (local_s, l + 1);
    }

  l = 0;

  for (a = 0; a < strlen (s); a++)
    {
      if (s[a] == '\'')
	local_s[l++] = '\\';
      if (s[a] == '\\')
	local_s[l++] = '\\';
      local_s[l++] = s[a];
    }

  local_s[l] = 0;
  return local_s;
}


char *
escape_pipe (char *s)
{
  static char *local_s = 0;
  static int local_s_len = 0;
  char *ptr;
  int a;
  int l;
  ptr = strchr (s, '|');
  if (ptr == 0)
    return s;
  l = 0;
  for (a = 0; a < strlen (s); a++)
    {
      if (s[a] == '\'')
	l++;
      l++;
    }

  if (l > local_s_len)
    {
      local_s_len = l;
      local_s = realloc (local_s, l + 10);
    }

  l = 0;

  for (a = 0; a < strlen (s); a++)
    {
      if (s[a] == '|')
	local_s[l++] = '\\';
      local_s[l++] = s[a];
    }
  local_s[l] = 0;
  return local_s;
}





void print_string (char *name, char *val, int l)
{
  static char *buff = 0;
  static int buff_len = 0;

  if (buff_len < l)
    {
      buff = realloc (buff, l + 1);
      buff_len = l;
    }

  memset (buff, 0, l + 1);
  memcpy (buff, val, l);
  buff[l + 1] = 0;
  printf ("%-18s: %s\n", name, buff);

}



void print_systables ()
{
  print_string ("TABNAME", tabrecord.tabname, 18);
  print_string ("OWNER", tabrecord.owner, 8);
#ifdef V7
  print_string ("DIRPATH", tabrecord.dirpath, 256);
#else
  print_string ("DIRPATH", tabrecord.dirpath, 64);
#endif
  print_long ("TABID", tabrecord.tabid);
  print_short ("ROWSIZE", tabrecord.rowsize);
printf("%x\n", tabrecord.rowsize);
  print_short ("NCOLS", tabrecord.ncols);
printf("%x\n", tabrecord.ncols);
  print_short ("NINDEXES", tabrecord.nindexes);
  print_long ("NROWS", tabrecord.nrows);
  print_long ("CREATED", tabrecord.created);
  print_long ("VERSION", tabrecord.version);
  print_string ("TABTYPE", tabrecord.tabtype, 1);
#ifdef V7
  print_string ("AUDPATH", tabrecord.audpath, 256);
#else
  print_string ("AUDPATH", tabrecord.audpath, 64);
#endif
  print_hex ("STAT", tabrecord.stat);
  printf ("---------------------------------------------\n");
}



void
A4GL_trim (char *s)
{
  int a;
  int has_len;
  if (strlen (s))
    has_len = 1;
  else
    has_len = 0;

  if (s == 0)
    return;
  if (s[0] == 0)
    return;

  for (a = strlen (s) - 1; a >= 0; a--)
    {
      if (s[a] == ' ')
	s[a] = 0;
      else
	break;
    }

  if (strlen (s) == 0 && has_len)
    {
      s[0] = ' ';
    }

}

int
read_column_definitions_from_db (int tabid, int ncols,
				 struct s_column **colptr, int rowsize)
{
  int thisize;
  struct s_column *columns;
  struct s_colrecord colrecord;
  int a;
  int offset;

  // Allocate enough space for ALL our columns in this table
  columns = malloc (sizeof (struct s_column) * ncols);
  *colptr = columns;

  // Clear everything down..
  memset (columns, 0, 0);


  // We need some way to check that we've read in all columns - 
  // so set the colno to -1 , when it gets read in this will be
  // replaced with the real column number...

  for (a = 0; a < ncols; a++)
    {
      columns[a].colno = -1;
    }


  // We'll start at the beginning of the file..
  rewind (syscolumns);

  // And go through each reacord...
  while (!feof (syscolumns))
    {
      short colno;
      int ctabid;

      // Read in a column - could be from any table
      fread (&colrecord, sizeof (colrecord), 1, syscolumns);
      if (colrecord.stat == 0)
	continue;		// Deleted...

      // Get our tabid from the table
      ctabid = ntohl (colrecord.tabid);
      // Is it the one we are looking for ?
      if (ctabid != tabid)
	continue;

      // OK - whats the column number ?
      colno = a_ntohs (colrecord.colno);
      if (verbose) {
      		printf ("colno=%d\n", colno);
	}
      // Copy it into our column list...


      memcpy (columns[colno - 1].colname, colrecord.colname, 18);
      columns[colno - 1].colname[18] = 0;
      A4GL_trim (columns[colno - 1].colname);
	//printf("%d - %s\n",colno, columns[colno - 1].colname);
      columns[colno - 1].tabid = tabid;
      columns[colno - 1].colno = colno;
      columns[colno - 1].coltype = a_ntohs (colrecord.coltype) & 0xff;
      columns[colno - 1].collength = a_ntohs (colrecord.collength);
      columns[colno - 1].offset = 0;
    }



  // Now check that all those -1 records have been changed
  //
  for (a = 0; a < ncols; a++)
    {
      if (columns[a].colno == -1)
	{
	  // We appear to be missing one...
	  fprintf
	    (stderr,"Unable to process - missing column definition for column %d\n", a);
	  exit (0);
	  return 0;
	}
    }

  offset = 0;
	if (show_schema ) {
		openOutFile();
		char buff[200];
	      	strncpy (buff, tabrecord.tabname, 18);
		buff[18]=0;
		A4GL_trim(buff);
		if (mode == 'I' || mode=='i' || mode=='P') {
			fprintf(out, "-- DROP TABLE %s;\n",  insert_tablename);
			fprintf(out, "CREATE TABLE %s (\n",  insert_tablename);
		} else {
			fprintf(out,"-- DROP TABLE %s;\n",  buff);
			fprintf(out,"CREATE TABLE %s (\n",  buff);
		}
  		for (a = 0; a < ncols; a++) {
			if (a<ncols-1) {
				fprintf(out,"   %s %s,\n", columns[a].colname, get_col_desc(columns[a].coltype, columns[a].collength));
			} else {
				fprintf(out,"   %s %s\n", columns[a].colname, get_col_desc(columns[a].coltype, columns[a].collength));
			}
		}
		fprintf(out,");\n\n");

	}
  // We've got to here - we appear to be ok...
  for (a = 0; a < ncols; a++)
    {
      columns[a].offset = offset;

      thisize = acl_rtypesize (columns[a].coltype, columns[a].collength);
      if (verbose)
	{
	  print_string ("CCOLNAME", columns[a].colname, 18);
	  print_long ("CTABID", ntohl (columns[a].tabid));
	  print_short ("CCOLNO", columns[a].colno);
	  print_short ("CCOLTYPE", columns[a].coltype);
	  print_short ("CCOLLENGTH", columns[a].collength);
	  printf (" --> %d\n", thisize);
	  printf ("---------------------------\n");
	}

      offset += thisize;
    }


  if (rowsize != offset)
    {
      printf ("Mismatch - expecting rowsize of %d - got %d\n", rowsize,
	      offset);
      return 0;
    }
  return 1;

}






int
read_table (struct s_column *columns, int ncols, int rowsize, char *dirpath)
{
  FILE *table;
  char *databuffer;
  int offset;
  int a;
  int max_cnt;
  table = fopen (dirpath, "r");
  if (table == 0)
    {
      printf ("Unable to open the .dat file '%s'\n", dirpath);
      return 0;
    }
  fseek(table,0,SEEK_END);
  max_cnt=ftell(table);
  progress_max_cnt=max_cnt/rowsize+1;
  progress_cnt=0;
  rewind (table);
  databuffer = malloc (rowsize + 1);



  if (verbose)
    {
      printf ("READING TABLE : %s (%p)\n", dirpath, table);
    }

  while (!feof (table))
    {
      int rs;
      memset (databuffer, 0, rowsize + 1);
      rs = fread (databuffer, rowsize + 1, 1, table);
      if (rs != 1)
	{
	  break;
	}
      if (databuffer[rowsize] == 0)
	{
	  //printf("Deleted\n");
	  continue;		// Its a deleted row...
	}

      offset = 0;

	progress_cnt++;
	print_progress(0);


      if (mode == 'S')
	{
	  for (a = 0; a < ncols; a++)
	    {
	      offset = columns[a].offset;
	      fprintf (out, "(%04d) %-18s ", offset, columns[a].colname);
	      fprintf (out, "%s\n",
		       dump_data (&databuffer[offset], columns[a].coltype,
				  columns[a].collength));
	    }
	  fprintf (out, "\n");
	}

      if (mode == 'U')
	{
	  for (a = 0; a < ncols; a++)
	    {
	      offset = columns[a].offset;
	      fprintf (out, "%s|",
		       escape_pipe (dump_data
				    (&databuffer[offset], columns[a].coltype,
				     columns[a].collength)));
	    }
	  fprintf (out, "\n");
	}

      if (mode == 'u')
	{
	  for (a = 0; a < ncols; a++)
	    {
	      offset = columns[a].offset;
	      fprintf (out, "%s|",
		       escape_pipe (dump_data
				    (&databuffer[offset], columns[a].coltype,
				     columns[a].collength)));
	    }
	  fprintf (out, "\n");
	}


      if (mode == 'i')
	{
	  fprintf (out, "iNSERT INTO  %s VALUES (\n   ", insert_tablename);
	  for (a = 0; a < ncols; a++)
	    {
	      char *b;
	      offset = columns[a].offset;

	      b =
		dump_data (&databuffer[offset], columns[a].coltype,
			   columns[a].collength);

	      if (strlen (b) == 0)
		{
		  printf ("NULL");
		}
	      else
		{

		      if (needs_quotes (columns[a].coltype))
			{
			  fprintf (out, "'");
			}
		      fprintf (out, "%s", escape_single_quotes (b));
		      if (needs_quotes (columns[a].coltype))
			{
			  fprintf (out, "'");
			}

		}
		  if (a < ncols - 1)
		    {
		      fprintf (out, ",\n   ");
		    }
	    }
	  fprintf (out, "\n);\n\n");
	}

      if (mode == 'I' || mode == 'P')
	{
	  char buff[120000];

	  sprintf (buff, "INSERT INTO  %s (", insert_tablename);
	  for (a = 0; a < ncols; a++)
	    {
	      if (a)
		strcat (buff, ",");
	      strcat (buff, columns[a].colname);
	    }
	  strcat (buff, ") VALUES (\n   ");
	  for (a = 0; a < ncols; a++)
	    {
	      char *b;
	      offset = columns[a].offset;

	      b =
		dump_data (&databuffer[offset], columns[a].coltype,
			   columns[a].collength);

	      if (strlen (b) == 0)
		{
		  strcat (buff, "NULL");
		}
	      else
		{
	      if (needs_quotes (columns[a].coltype)) { strcat (buff, "'"); }
		  strcat (buff,
			  escape_single_quotes (dump_data
						(&databuffer[offset],
						 columns[a].coltype,
						 columns[a].collength)));
	      if (needs_quotes (columns[a].coltype)) { strcat (buff, "'"); }
		}

	      if (a < ncols - 1)
		{
		  if (mode == 'P')
		    {
		      strcat (buff, ",");
		    }
		  else
		    {
		      strcat (buff, ",\n   ");
		    }
		}
	    }

	  if (mode == 'P')
	    {
	      strcat (buff, ");");

	      if (verbose)
		{
		  printf ("Executing : %s\n", buff);
		}

	      if (!DBIO_execute_sql (buff))
		{
		  fprintf (stderr,"Failed to execute : '%s'\n",buff);
		  exit (2);
		}
	    }
	  else
	    {
	      strcat (buff, "\n);\n\n");
	      fprintf (out, "%s\n", buff);
	    }
	}



    }


  print_progress(1);
  return 1;
}



static int needs_quotes (int n)
{
  if (n == 0 || n == 7 || n == 10)
    return 1;
  else
    return 0;
}



int main (int argc, char *argv[])
{
  int tabid;
  struct s_column *columns = 0;
  int ncolumns = 0;
  int rowsize = 0;
  int c;
  int ok;
  datefmt = datefmt_s;

  if (argc == 1)
    {
      usage (argv[0], 0);
    }



  ok = process_argv (argc, argv);
  if (ok == 0)
    {
      usage (argv[0], 0);
    }
  if (ok < 0)
    {
      usage (argv[0], 1);
    }


  if (verbose)
    {
      printf ("Search : Name = '%s' And/Or ID=%d\n", search_tablename,
	      search_tabid);
    }

  if (strlen (search_tablename) || search_tabid)
    {
      int read_first = 0;
      int found_something = 0;
      /* Open up our two file handles.. */
      if (verbose)
	{
	  printf ("Have a table to find\n");
	}
      systables = fopen ("systables.dat", "r");
      syscolumns = fopen ("syscolumns.dat", "r");

      if (systables == 0)
	{
	  fprintf (stderr,"Unable to open systables.dat\n");
	  exit (1);
	}
      if (syscolumns == 0)
	{
	  fprintf (stderr,"Unable to open syscolumns.dat\n");
	  exit (1);
	}
      if (verbose)
	{
	  printf ("Opened systables.dat and syscolumns.dat\n");
	}





      /* Go through each record in the systables table */
      while (!feof (systables))
	{
	  char dirpath[512];
	  if (verbose)
	    {
	      printf ("Reading a record for a table\n");
	    }
	  fread (&tabrecord, sizeof (tabrecord), 1, systables);

	  if (tabrecord.stat == 0)
	    continue;		// Deleted...

	  // Work out the real tabid
	  tabid = ntohl (tabrecord.tabid);

	  if (read_first == 0)
	    {
	      read_first = 1;
	      if (verbose)
		{
		  printf ("Tabid =%d for first table\n", tabid);
		}
	      if (tabid != 1)
		{
#ifdef V7
		  fprintf
		    (stderr,"Possibly the wrong version ? - recompile without using -D V7\n");
#else
		  fprintf
		    (stderr,"Possibly the wrong version ? - recompile using -D V7\n");
#endif
		  exit (2);
		}
	    }
	  if (verbose)
	    {
	      printf ("Tabid=%d\n", tabid);
	    }


	  if ((tabid != search_tabid) && search_tabid > 0)
	    {
	      if (verbose)
		{
		  printf ("Using tabids - %d != %d\n", tabid, search_tabid);
		}
	      continue;
	    }
	  if (search_tabid == 0)
	    {
	      char buff1[19];
	      char buff2[19];
	      strncpy (buff1, tabrecord.tabname, 18);
	      strncpy (buff2, search_tablename, 18);
	      buff1[18] = 0;
	      buff2[18] = 0;
	      A4GL_trim (buff1);
	      A4GL_trim (buff2);
	      if (strcmp (buff1, buff2) != 0)
		{
		  if (verbose)
		    {
		      printf ("'%s' '%s'\n", buff1, buff2);
		    }
		  continue;
		}

	    }

	  found_something = 1;
	  if (verbose)
	    {
	      printf ("Found my table\n");
	    }
	  // Generate the filename of the data for this table...
	  memset (dirpath, 0, sizeof (dirpath));
	  strcpy (dirpath, "./");
	  strncat (dirpath, tabrecord.dirpath, 256);
	  A4GL_trim (dirpath);
	  strcat (dirpath, ".dat");

	  strcpy (datafile, dirpath);

	  if (verbose)
	    {
	      printf ("Think datafile is called : %s\n", datafile);
	    }
	  // Read in the column list...
	  ncolumns = a_ntohs (tabrecord.ncols);
	  rowsize = a_ntohs (tabrecord.rowsize);


	  if (verbose)
	    {
	      printf ("ncolumns=%d rowsize=%d\n", ncolumns, rowsize);
	    }


	  if (!read_column_definitions_from_db
	      (tabid, ncolumns, &columns, a_ntohs (tabrecord.rowsize)))
	    {
	      fprintf (stderr,"Unable to read columns\n");
	      exit (2);
	    }


	  break;		// We've found it - so we don't need to do any more....
	}
      if (found_something == 0)
	{
	  fprintf (stderr,"Couldn't find the specified table\n");
	}
    }

  if (strlen (columnfile))
    {
      if (verbose)
	{
	  printf ("Column file specified\n");
	}
      if (strlen (search_tablename) || search_tabid)
	{
	  FILE *fout;
	  if (verbose)
	    {
	      printf
		("We have -t or -T - so we want to write the definition\n");
	    }
	  if (columns == 0)
	    {
	      printf ("No column definitions found to write\n");
	    }

	  fout = fopen (columnfile, "w");

	  if (fout == 0)
	    {
	      fprintf (stderr,"Unable to write column definition file\n");
	      exit (2);
	    }

	  fwrite (&rowsize, 4, 1, fout);
	  fwrite (columns, a_ntohs (tabrecord.ncols) * sizeof (struct s_column),
		  1, fout);
	  fclose (fout);
	  if (verbose)
	    {
	      printf ("Written column info to %s\n", columnfile);
	    }
	  exit (0);
	}
      else
	{
	  // We want to read these in...
	  FILE *fin;
	  int n;
	  if (verbose)
	    {
	      printf
		("We dont have -t or -T - so we want to read the definition\n");
	    }
	  fin = fopen (columnfile, "r");
	  if (fin == 0)
	    {
	      fprintf (stderr,"Unable to read column definition file\n");
	      exit (2);
	    }
	  fseek (fin, 0, SEEK_END);
	  n = ftell (fin) - 4;
	  rewind (fin);

	  if (n % sizeof (struct s_column) != 0)
	    {
	      fprintf (stderr,"Invalid columns file\n");
	      exit (1);
	    }

	  columns = malloc (n + 4);
	  fread (&rowsize, 4, 1, fin);
	  fread (columns, n + 4, 1, fin);

	  n = n / sizeof (struct s_column);
	  ncolumns = n;
	  fclose (fin);
	}
    }


  openOutFile();


  if (columns && ncolumns && rowsize && strlen (datafile))
    {
      if (mode == 'P'||mode=='u')
	{
	  if (!DBIO_connect_db (dbname, uname, passwd))
	    {
	      printf ("Unable to open database\n");
	    }
	  if (clear_first)
	    {
	      if (!DBIO_clear_table (insert_tablename))
		{
		  printf ("Unable to clear table\n");
		}
	    }
	}

      if (read_table (columns, ncolumns, rowsize, datafile))
	{
	  exit (0);
	}
      else
	{
	  fprintf (stderr,"Some error\n");
	  exit (1);
	}
    }

return 0;
}







int
acl_rtypesize (int dtype, int len)
{
#define DECLEN(m,n)     (((m)+((n)&1)+3)/2)
#define DECLENGTH(len)  DECLEN(PRECTOT(len),PRECDEC(len))
#define PRECTOT(x)      (((x)>>8) & 0xff)
#define PRECDEC(x)      ((x) & 0xff)
#define PRECMAKE(x,y)   (((x)<<8) + (y))

  switch (dtype & 0xf)
    {
    case 1:
      return 2;;		// SMALLINT
    case 2:
      return 4;
      break;			// INTEGER
    case 3:
      return 8;
      break;			// FLOAT
    case 4:
      return 4;
      break;			// SMFLOAT
    case 6:
      return 4;
      break;			// SERIAL
    case 7:
      return 4;
      break;			// DATE

// Dependant on collength
    case 0:
      return len;		// CHAR (n)
    case 5:
      return DECLENGTH (len);	// DECIMAL
    case 8:
      return DECLENGTH (len);	// MONEY

    case 10:
      return ((len >> 8) / 2) + 1;
      break;			// DTIME



      // Unexpected...
    case 14:			// INTERVAL
    case 9:			// NULL
    case 11:			// BYTES
    case 12:			// TEXT
    case 13:			// VARCHAR
    case 15:			// NCHAR
    default:
      fprintf (stderr,"Unexpected datatype: %d\n", dtype);
      exit (2);
    }
}

int check_datefmt ()
{
  int a;
  for (a = 0; a < strlen (datefmt); a++)
    {
      datefmt[a] = toupper (datefmt[a]);
    }
  if (strcmp (datefmt, "DMY4/") == 0)
    return 1;
  if (strcmp (datefmt, "DMY4-") == 0)
    return 1;
  if (strcmp (datefmt, "DMY2/") == 0)
    return 1;
  if (strcmp (datefmt, "DMY2-") == 0)
    return 1;

  if (strcmp (datefmt, "MDY4/") == 0)
    return 1;
  if (strcmp (datefmt, "MDY4-") == 0)
    return 1;
  if (strcmp (datefmt, "MDY2/") == 0)
    return 1;
  if (strcmp (datefmt, "MDY2-") == 0)
    return 1;

  if (strcmp (datefmt, "YMD4/") == 0)
    return 1;
  if (strcmp (datefmt, "YMD4-") == 0)
    return 1;
  if (strcmp (datefmt, "YMD2/") == 0)
    return 1;
  if (strcmp (datefmt, "YMD2-") == 0)
    return 1;

  printf ("Unknown date format. I know about:\n\n");
  printf ("   DMY4/ DMY4- DMY2/ DMY2-\n");
  printf ("   MDY4/ MDY4- MDY2/ MDY2-\n");
  printf ("   YMD4/ YMD4- YMD2/ YMD2-\n\n");
  if (!strchr (datefmt, '-') && !strchr (datefmt, '/'))
    {
      printf ("Don't forget the separator - it doesn't default to a '/'!\n");
    }
  printf ("\n\n");


  exit (1);
}


static void print_progress(int finished) {
static int ldisplay=-1;
if (progress_type=='N') {
		if (finished) chk_unl(finished);
		return;
}
if (progress_type=='%') {
		if (!finished) {
			int pc;
			pc=(progress_cnt*100)/progress_max_cnt;
			if (pc!=ldisplay) {ldisplay=pc; printf("\r%d %%  ",pc); fflush(stdout); chk_unl(finished);}
		} else {
			printf("100 %%\n\n");
		}
}


if (progress_type == '1')
  {
    if (!finished)
      {
	if (progress_cnt % 10 == 0)
	  {
	    printf ("\r%d", progress_cnt);
	    fflush (stdout);
	    chk_unl (finished);
	  }
      }
    else
      {
	printf ("\r%d\n", progress_cnt);
	chk_unl (finished);
      }
  }
if (progress_type == '2')
  {
    if (!finished)
      {
	if (progress_cnt % 100 == 0)
	  {
	    printf ("\r%d", progress_cnt);
	    fflush (stdout);
	    chk_unl (finished);
	  }
      }
    else
      {
	printf ("\r%d\n", progress_cnt);
	chk_unl (finished);
      }
  }
if (progress_type == '3')
  {
    if (!finished)
      {
	if (progress_cnt % 1000 == 0)
	  {
	    printf ("\r%d", progress_cnt);
	    fflush (stdout);
	    chk_unl (finished);
	  }
      }
    else
      {
	printf ("\r%d\n", progress_cnt);
	chk_unl (finished);
      }
  }
if (progress_type == '4')
  {
    if (!finished)
      {
	if (progress_cnt % 10000 == 0)
	  {
	    printf ("\r%d", progress_cnt);
	    fflush (stdout);
	    chk_unl (finished);
	  }
      }
    else
      {
	printf ("\r%d\n", progress_cnt);
	chk_unl (finished);
      }
  }
if (progress_type == '5')
  {
    if (!finished)
      {
	if (progress_cnt % 100000 == 0)
	  {
	    printf ("\r%d", progress_cnt);
	    fflush (stdout);
	    chk_unl (finished);
	  }
      }
    else
      {
	printf ("\r%d\n", progress_cnt);
	chk_unl (finished);
      }
  }
if (progress_type == '6')
  {
    if (!finished)
      {
	if (progress_cnt % 1000000 == 0)
	  {
	    printf ("\r%d", progress_cnt);
	    fflush (stdout);
	    chk_unl (finished);
	  }
      }
    else
      {
	printf ("\r%d\n", progress_cnt);
	chk_unl (finished);
      }
  }
}



void chk_unl(int finished)  {
if (mode=='u') {
	if (verbose>4) { printf("Closing\n"); }
	fclose(out);
	if (verbose>2) { printf("Loading\n"); }
	DBIO_load_unl(insert_tablename,outfile);
	if (!finished) {
		out=fopen(outfile,"w");
		if (verbose>2) { printf("Re-opening file\n"); }
	}
}
}

static char *decode_decimal(int a) {
int m;
static char buff[200];
        m=a% 256;
        a=(a-m)/256;


        if (m==255) 
		sprintf(buff,"%d",a);
        else
		sprintf(buff,"%d,%d",a,m);

	return buff;
}


static char *decode_datetime(int a) {
static char lv_str[200];
int pt1,pt2 ;
char lv_part1[30];
char lv_part2[30];
int fr2 ;

if (((a/16) % 16) <= 10) {
         pt1=(((a / 16) % 16) / 2);
} else {
        pt1=6;
}

strcpy(lv_part1,dtparts[pt1+1]);



if ((a %16) <= 10 ) {
        pt2= ((a % 16) / 2);
        strcpy(lv_part2,dtparts[pt2+1] );
} else {
        pt2=6;
        fr2 = (a % 16) - 10;
        sprintf(lv_part2,"%s(%d)",dtparts[pt2+1] ,fr2);
}

sprintf(lv_str, "%s TO %s", lv_part1 ,lv_part2);

return lv_str;
}


static char *decode_interval(int a) {
int p1;
static char lv_str[200];
char lv_part1[200];
char lv_part2[200];
int pt1;
int pt2;
int p2;
int fr1;
int fr2;
         p1 = (a / 16) % 16;
        if (p1 <= 10) pt1=  (p1 / 2);
        else pt1= 6;

        fr2 = (a % 16) - 10;

        p2 = a % 16;

        if (p2 <= 10) 
                pt2= (p2 / 2);
        else
                pt2= 6;

        if (pt1 != 6) {
                fr1 = (a / 256) + p1 - p2;
                if (fr1!=2)
                        sprintf( lv_part1,"%s(%d)",dtparts[pt1+1] ,fr1);
                else
                        sprintf( lv_part1,"%s(%d)",dtparts[pt1+1],fr1);
	} else {
                sprintf( lv_part1,"%s",dtparts[pt1+1]);
	}

        if (pt2 == 6)  {
                sprintf( lv_part2,"%s(%d)",dtparts[pt2+1],fr2 );
	} else {
                sprintf( lv_part2,"%s",dtparts[pt2+1]);
	}
	sprintf(lv_str,"INTERVAL %s TO %s", lv_part1 , lv_part2);
        return lv_str;
}


static char *decode_varchar(int a) {
static char lv_str[200];
if ((a / 256) %256) {
	sprintf(lv_str,"%d,%d", (a%256) ,((a / 256)%256));
} else {
	sprintf(lv_str,"%d", (a%256) );
}
return lv_str;
}

char *get_col_desc(int coltype,int collength) {
	static char coldesc[2000];
char *lv_nn;
               if (coltype>255){
                                coltype=coltype-256;
                                lv_nn=" NOT NULL";
			} else {
                                lv_nn="";
			}


                        switch(coltype) {
                                case 0 : sprintf(coldesc,"CHAR(%d)%s",collength ,lv_nn); break;
                                case 1 : sprintf(coldesc,"SMALLINT%s",lv_nn); break;
                                case 2 : sprintf(coldesc,"INTEGER%s",lv_nn); break;
                                case 3 : sprintf(coldesc,"FLOAT%s",lv_nn); break;
                                case 4 : sprintf(coldesc,"SMALLFLOAT%s",lv_nn); break;

                                case 5 : sprintf(coldesc,"DECIMAL(%s)%s",decode_decimal(collength),lv_nn); break;

                                case 6 : sprintf(coldesc,"SERIAL%s",lv_nn); break;
                                case 7 : sprintf(coldesc,"DATE%s",lv_nn); break;
                                case 8 : sprintf(coldesc,"MONEY(%s)%s",decode_decimal(collength),lv_nn); break;
                                case 9 : sprintf(coldesc,"---"); break;

                                case 10 : sprintf(coldesc,"DATETIME %s%s",decode_datetime(collength),lv_nn); break;
                                case 11 : sprintf(coldesc,"BYTE%s",lv_nn); break;
                                case 12 : sprintf(coldesc,"TEXT%s",lv_nn); break;
                                case 13 : sprintf(coldesc,"VARCHAR(%s)%s",decode_varchar(collength) ,lv_nn); break;
                                case 14 : sprintf(coldesc,"%s%s",decode_interval(collength) ,lv_nn); break;
                                case 15 : sprintf(coldesc,"NCHAR(%d)%s",collength,lv_nn); break;
                                default: printf("INVALID DATATYPE: %d\n",coltype); exit (1);
                        }

	return coldesc;
}
